@@ -1,5 +1,6 @@
 <?php
-// Icinga PDF Export | (c) 2019 Icinga GmbH | GPLv2
+
+/* Icinga PDF Export | (c) 2019 Icinga GmbH | GPLv2 */
 
 namespace Icinga\Module\Pdfexport\Controllers;
 
@@ -16,14 +17,14 @@ class ConfigController extends Controller
         parent::init();
     }
 
-    public function binaryAction()
+    public function chromeAction()
     {
         $form = (new ChromeBinaryForm())
             ->setIniConfig(Config::module('pdfexport'));
 
         $form->handleRequest();
 
-        $this->view->tabs = $this->Module()->getConfigTabs()->activate('binary');
+        $this->view->tabs = $this->Module()->getConfigTabs()->activate('chrome');
         $this->view->form = $form;
     }
 }
