@@ -12,8 +12,11 @@ use Icinga\File\Storage\TemporaryLocalFileStorage;
 use ipl\Html\HtmlString;
 use LogicException;
 use React\ChildProcess\Process;
-use React\EventLoop\Factory;
+use React\EventLoop\Loop;
 use React\EventLoop\TimerInterface;
+use React\Promise;
+use React\Promise\ExtendedPromiseInterface;
+use Throwable;
 use WebSocket\Client;
 use WebSocket\ConnectionException;
 
@@ -240,26 +243,33 @@ JS;
     }
 
     /**
-     * Export to PDF
+     * Generate a PDF raw string asynchronously.
      *
-     * @return string
-     * @throws Exception
+     * @return ExtendedPromiseInterface
      */
-    public function toPdf()
+    public function asyncToPdf(): ExtendedPromiseInterface
     {
-        switch (true) {
-            case $this->remote !== null:
-                try {
-                    $result = $this->jsonVersion($this->remote[0], $this->remote[1]);
-                    $parts = explode('/', $result['webSocketDebuggerUrl']);
-                    $pdf = $this->printToPDF(join(':', $this->remote), end($parts), isset($this->document)
-                        ? $this->document->getPrintParameters()
-                        : []);
-                    break;
-                } catch (Exception $e) {
-                    if ($this->binary === null) {
-                        throw $e;
-                    } else {
+        $deferred = new Promise\Deferred();
+        Loop::futureTick(function () use ($deferred) {
+            switch (true) {
+                case $this->remote !== null:
+                    try {
+                        $result = $this->jsonVersion($this->remote[0], $this->remote[1]);
+                        if (is_array($result)) {
+                            $parts = explode('/', $result['webSocketDebuggerUrl']);
+                            $pdf = $this->printToPDF(
+                                join(':', $this->remote),
+                                end($parts),
+                                ! $this->document->isEmpty() ? $this->document->getPrintParameters() : []
+                            );
+                            break;
+                        }
+                    } catch (Exception $e) {
+                        if ($this->binary == null) {
+                            $deferred->reject($e);
+                            return;
+                        }
+
                         Logger::warning(
                             'Failed to connect to remote chrome: %s:%d (%s)',
                             $this->remote[0],
@@ -267,80 +277,134 @@ JS;
                             $e
                         );
                     }
-                }
+
+                    // Reject the promise if we didn't get the expected output from the /json/version endpoint.
+                    if ($this->binary === null) {
+                        $deferred->reject(
+                            new Exception('Failed to determine remote chrome version via the /json/version endpoint.')
+                        );
+                        return;
+                    }
 
                 // Fallback to the local binary if a remote chrome is unavailable
-            case $this->binary !== null:
-                $browserHome = $this->getFileStorage()->resolvePath('HOME');
-                $commandLine = join(' ', [
-                    escapeshellarg($this->getBinary()),
-                    static::renderArgumentList([
-                        '--bwsi',
-                        '--headless',
-                        '--disable-gpu',
-                        '--no-sandbox',
-                        '--no-first-run',
-                        '--disable-dev-shm-usage',
-                        '--remote-debugging-port=0',
-                        '--homedir=' => $browserHome,
-                        '--user-data-dir=' => $browserHome
-                    ])
-                ]);
+                case $this->binary !== null:
+                    $browserHome = $this->getFileStorage()->resolvePath('HOME');
+                    $commandLine = join(' ', [
+                        escapeshellarg($this->getBinary()),
+                        static::renderArgumentList([
+                            '--bwsi',
+                            '--headless',
+                            '--disable-gpu',
+                            '--no-sandbox',
+                            '--no-first-run',
+                            '--disable-dev-shm-usage',
+                            '--remote-debugging-port=0',
+                            '--homedir=' => $browserHome,
+                            '--user-data-dir=' => $browserHome
+                        ])
+                    ]);
 
-                if (Platform::isLinux()) {
-                    Logger::debug('Starting browser process: HOME=%s exec %s', $browserHome, $commandLine);
-                    $chrome = new Process('exec ' . $commandLine, null, ['HOME' => $browserHome]);
-                } else {
-                    Logger::debug('Starting browser process: %s', $commandLine);
-                    $chrome = new Process($commandLine);
-                }
-
-                $loop = Factory::create();
-
-                $killer = $loop->addTimer(10, function (TimerInterface $timer) use ($chrome) {
-                    $chrome->terminate(6); // SIGABRT
-                    Logger::error(
-                        'Terminated browser process after %d seconds elapsed without the expected output',
-                        $timer->getInterval()
-                    );
-                });
-
-                $chrome->start($loop);
-
-                $pdf = null;
-                $chrome->stderr->on('data', function ($chunk) use (&$pdf, $chrome, $loop, $killer) {
-                    Logger::debug('Caught browser output: %s', $chunk);
-
-                    if (preg_match(self::DEBUG_ADDR_PATTERN, trim($chunk), $matches)) {
-                        $loop->cancelTimer($killer);
-
-                        try {
-                            $pdf = $this->printToPDF($matches[1], $matches[2], isset($this->document)
-                                ? $this->document->getPrintParameters()
-                                : []);
-                        } catch (Exception $e) {
-                            Logger::error('Failed to print PDF. An error occurred: %s', $e);
-                        }
-
-                        $chrome->terminate();
+                    if (Platform::isLinux()) {
+                        Logger::debug('Starting browser process: HOME=%s exec %s', $browserHome, $commandLine);
+                        $chrome = new Process('exec ' . $commandLine, null, ['HOME' => $browserHome]);
+                    } else {
+                        Logger::debug('Starting browser process: %s', $commandLine);
+                        $chrome = new Process($commandLine);
                     }
-                });
 
-                $chrome->on('exit', function ($exitCode, $termSignal) use ($loop, $killer) {
-                    $loop->cancelTimer($killer);
+                    $killer = Loop::addTimer(10, function (TimerInterface $timer) use ($chrome, $deferred) {
+                        $chrome->terminate(6); // SIGABRT
 
-                    Logger::debug('Browser terminated by signal %d and exited with code %d', $termSignal, $exitCode);
-                });
+                        Logger::error(
+                            'Browser timed out after %d seconds without the expected output',
+                            $timer->getInterval()
+                        );
 
-                $loop->run();
-        }
+                        $deferred->reject(
+                            new Exception(
+                                'Received empty response or none at all from browser.'
+                                . ' Please check the logs for further details.'
+                            )
+                        );
+                    });
 
-        if (empty($pdf)) {
-            throw new Exception(
-                'Received empty response or none at all from browser.'
-                . ' Please check the logs for further details.'
-            );
-        }
+                    $chrome->start();
+
+                    $chrome->stderr->on('data', function ($chunk) use ($chrome, $deferred, $killer) {
+                        Logger::debug('Caught browser output: %s', $chunk);
+
+                        if (preg_match(self::DEBUG_ADDR_PATTERN, trim($chunk), $matches)) {
+                            Loop::cancelTimer($killer);
+
+                            try {
+                                $pdf = $this->printToPDF(
+                                    $matches[1],
+                                    $matches[2],
+                                    ! $this->document->isEmpty() ? $this->document->getPrintParameters() : []
+                                );
+                            } catch (Exception $e) {
+                                Logger::error('Failed to print PDF. An error occurred: %s', $e);
+                            }
+
+                            $chrome->terminate();
+
+                            if (! empty($pdf)) {
+                                $deferred->resolve($pdf);
+                            } else {
+                                $deferred->reject(
+                                    new Exception(
+                                        'Received empty response or none at all from browser.'
+                                        . ' Please check the logs for further details.'
+                                    )
+                                );
+                            }
+                        }
+                    });
+
+                    $chrome->on('exit', function ($exitCode, $signal) use ($killer) {
+                        Loop::cancelTimer($killer);
+
+                        Logger::debug('Browser terminated by signal %d and exited with code %d', $signal, $exitCode);
+
+                        // Browser is either timed out (after 10s) and the promise should have already been rejected,
+                        // or it is terminated using its terminate() method, in which case the promise is also already
+                        // resolved/rejected. So, we don't need to resolve/reject the promise here.
+                    });
+
+                    return;
+            }
+
+            if (! empty($pdf)) {
+                $deferred->resolve($pdf);
+            } else {
+                $deferred->reject(
+                    new Exception(
+                        'Received empty response or none at all from browser.'
+                        . ' Please check the logs for further details.'
+                    )
+                );
+            }
+        });
+
+        return $deferred->promise();
+    }
+
+    /**
+     * Export to PDF
+     *
+     * @return string
+     * @throws Exception
+     */
+    public function toPdf()
+    {
+        $pdf = '';
+        // We don't intend to register any then/otherwise handlers, so call done on that promise
+        // to properly propagate unhandled exceptions to the caller.
+        $this->asyncToPdf()->done(function (string $newPdf) use (&$pdf) {
+            $pdf = $newPdf;
+        });
+
+        Loop::run();
 
         return $pdf;
     }
@@ -403,7 +467,7 @@ JS;
 
             // wait for page to fully load
             $this->waitFor($page, 'Page.frameStoppedLoading', ['frameId' => $frameId]);
-        } elseif (isset($this->document)) {
+        } elseif (! $this->document->isEmpty()) {
             // If there's no url to load transfer the document's content directly
             $this->communicate($page, 'Page.setDocumentContent', [
                 'frameId'   => $targetId,
@@ -420,7 +484,7 @@ JS;
         $this->waitFor($page, self::WAIT_FOR_NETWORK);
 
         // Wait for layout to initialize
-        if (isset($this->document)) {
+        if (! $this->document->isEmpty()) {
             // Ensure layout scripts work in the same environment as the pdf printing itself
             $this->communicate($page, 'Emulation.setEmulatedMedia', ['media' => 'print']);
 
@@ -588,6 +652,7 @@ JS;
         $wait = true;
         $interceptedPos = -1;
 
+        $params = null;
         do {
             if (isset($this->interceptedEvents[++$interceptedPos])) {
                 $response = $this->interceptedEvents[$interceptedPos];
